#include "range_tree.h"



static int
compare_ind_entries(const void * a_ptr, const void * b_ptr)
{
    double a = ((index_entry *)a_ptr)->value;
    double b = ((index_entry *)b_ptr)->value;

    if(isnan(a) && isnan(b)) return 0;
    if(isnan(a)) return -1;
    if(isnan(b)) return 1;
    if(a < b) return -1;
    if(a == b) return 0;
    return 1;
}

index_entry **
range_tree_create_indices(dataset * d)
{
    index_entry ** indices;
    size_t i, j;

    indices = malloc(sizeof(index_entry *) * d->n_items);
    if(indices != NULL)
    {
	// allocate memory
	for(i = 0; i < d->n_items; i++)
	{
	    indices[i] = malloc(d->n * sizeof(index_entry));
	    if(indices[i] == NULL)
	    {
		for(j = 0; j < i; j++)
		{
		    free(indices[i]);
		}
		free(indices);
		return NULL;
	    }
	}
	// copy data
	for(i = 0; i < d->n_items; i++)
	{
	    for(j = 0; j < d->n; j++)
	    {
		indices[i][j].record = d->data[j];
		indices[i][j].value = d->data[j][i];
	    }
	}
	// sort
	for(i = 0; i < d->n_items; i++)
	{
	    qsort(indices[i], d->n, sizeof(index_entry), compare_ind_entries);
	}
    }
    return indices;
}

void
range_tree_free_indices(index_entry ** indices, size_t n_indices)
{
    size_t i;

    for(i = 0; i < n_indices; i++)
    {
	free(indices[i]);
    }
    free(indices);
}




static void
range_tree_init(range_tree * r, double value)
{
    r->counter = 1;
    r->list = NULL;
    r->listlen = 0;
    r->value = value;
    //r->dim = ?;
}

range_tree *
range_tree_new(double value)
{
    range_tree * r = malloc(sizeof(range_tree));
    if(r != NULL)
    {
	range_tree_init(r, value);
    }
    return r;
}

static void
range_tree_free_rec(range_tree * r)
{
    size_t i;

    if(r->list != NULL)
    {
	for(i = 0; i < r->listlen; i++)
	{
	    range_tree_free_rec(r->list + i);
	}
	free(r->list);
    }
}

void
range_tree_free(range_tree * r)
{
    range_tree_free_rec(r);
    free(r);
}

static size_t 
compute_counts(range_tree * r, size_t beg, size_t n)
{
    size_t n2 = n / 2;
    size_t i = beg + n2;
    size_t tot, tot2;

    if(n == 0)
    {
	return 0;
    }
    tot = compute_counts(r, beg, n2);
    tot2 = compute_counts(r, i + 1, n - n2 - 1);
    r->list[i].counter += tot;
    return r->list[i].counter + tot2;
}

static void
add_dim_rec(range_tree * r, size_t a, index_entry * ind, size_t indlen,  size_t beg, size_t n)
{
    size_t n2 = n / 2;
    size_t i = beg + n2;
    size_t cnt1 = 0, cnt2 = 0, cnt3 = 0;
    size_t s, s1, s2, s3;
    index_entry * left, * left2, * right;

    if(n == 0)
    {
	return;
    }
    for(s = 0; s < indlen; s++)
    {
	if(!isnan(ind[s].record[r->dim]) && ind[s].record[r->dim] <  r->list[i].value) cnt1++;
	if(!isnan(ind[s].record[r->dim]) && ind[s].record[r->dim] <= r->list[i].value) cnt2++;
	if(!isnan(ind[s].record[r->dim]) && ind[s].record[r->dim] >  r->list[i].value) cnt3++;
    }
    left  = malloc(sizeof(index_entry) * cnt1);
    left2 = malloc(sizeof(index_entry) * cnt2);
    right = malloc(sizeof(index_entry) * cnt3);
    s1 = s2 = s3 = 0;
    for(s = 0; s < indlen; s++)
    {
	if(!isnan(ind[s].record[r->dim]) && ind[s].record[r->dim] <  r->list[i].value) left [s1++] = ind[s];
	if(!isnan(ind[s].record[r->dim]) && ind[s].record[r->dim] <= r->list[i].value) left2[s2++] = ind[s];
	if(!isnan(ind[s].record[r->dim]) && ind[s].record[r->dim] >  r->list[i].value) right[s3++] = ind[s];
    }
    range_tree_add_dim(r->list +i, a, left2, cnt2);
    if(n2 > 0)
	add_dim_rec(r, a, left,  cnt1, beg, n2);
    if(n > n2 + 1)
	add_dim_rec(r, a, right, cnt3, i + 1, n - n2 - 1);
    free(right);
    free(left2);
    free(left);
}


void range_tree_add_dim(range_tree * r, size_t a, index_entry * ind, size_t indlen)
{
    double x, old_x;
    size_t i;
    size_t cnt = 0;
    size_t counter;
    size_t s;

    if(indlen == 0)
	return;
    if(r->list == NULL)
    {
	r->dim = a;
	// count list size
	old_x = ind[0].value;
	cnt = 0;
	for(s = 0; s < indlen; s++)
	{
	    x = ind[s].value;
	    //if((isnan(old_x) && !isnan(x)) || x != old_x)
	    if(x != old_x)
	    {
		if(!isnan(old_x))
		{
		   cnt++;
		}
		old_x = x;
	    }
	}
	if(!isnan(old_x))
	{
	    cnt++;
	}
	r->listlen = cnt;

	// create the list
	r->list = malloc(sizeof(range_tree) * r->listlen);
	old_x = ind[0].value;
	cnt = 0;
	counter = 0;
	for(s = 0; s < indlen; s++)
	{
	    x = ind[s].value;
	    //if((isnan(old_x) && !isnan(x)) || x != old_x)
	    if(x != old_x)
	    {
		if(!isnan(old_x))
		{
		    range_tree_init(r->list + cnt, old_x);
		    r->list[cnt].counter = counter;
		    cnt++;
		}
		counter = 1;
		old_x = x;
	    }
	    else
	    {
		counter++;
	    }
	}
	if(!isnan(old_x))
	{
	    range_tree_init(r->list + cnt, old_x);
	    r->list[cnt].counter = counter;
	}

	compute_counts(r, 0, r->listlen);
    }
    else
    {
	add_dim_rec(r, a, ind, indlen, 0, r->listlen);
    }
}

static size_t
get_count_rec(range_tree * r, double * bounds, size_t lenbounds, size_t beg, size_t n)
{
    size_t n2 = n / 2;
    size_t i = beg + n2;
    double v;
    size_t ret = 0;

    if(n == 0)
    {
	return 0;
    }
    v = r->list[i].value;

    //if(!isnan(v) && !isnan(bounds[0]) && v < bounds[0])
    if(v < bounds[0])
    {
	ret = range_tree_get_count(r->list + i, bounds + 1, lenbounds - 1);
	ret += get_count_rec(r, bounds, lenbounds, i + 1, n - n2 - 1);
    }
    //else if(!isnan(v) && !isnan(bounds[0]) && v == bounds[0])
    else if(v == bounds[0])
    {
	ret = get_count_rec(r, bounds, lenbounds, beg, n2);
    }
    //else if(!isnan(v) && !isnan(bounds[0]) && v > bounds[0])
    else if(v > bounds[0])
    {
	ret = get_count_rec(r, bounds, lenbounds, beg, n2);
    }
    return ret;
}

size_t
range_tree_get_count(range_tree * r, double * bounds, size_t lenbounds)
{
    if(lenbounds == 0)
	return r->counter;
    return get_count_rec(r, bounds, lenbounds, 0, r->listlen);
}

void
range_tree_remove_dimension(range_tree * r, size_t n)
{
    size_t i;

    if(r->list == NULL)
    {
	return;
    }
    if(n == 2)
    {
	range_tree_free_rec(r);
	r->list = NULL;
    }
    else
    {
	for(i = 0; i < r->listlen; i++)
	{
	    range_tree_remove_dimension(r->list + i, n - 1);
	}
    }
}

